############
# External monitor setup
# Use this script to configure your monitoring for your GenAI app/agent deployed off Databricks.
# See the documentation for more details about the product capabilities: https://docs.databricks.com/aws/en/generative-ai/agent-evaluation/monitoring-non-agent-framework
# See the https://docs.databricks.com/aws/en/generative-ai/agent-evaluation/monitoring-non-agent-framework for a Databricks Notebook version.
############

import os
import mlflow
from mlflow.tracing.destination import Databricks
from databricks.sdk import WorkspaceClient

############
# Required dependencies:
#   mlflow>=2.21.2
#   databricks-agents>=0.18.1
#   databricks-connect # used to connect to Spark to create the monitoring tables
#   databricks-sdk[openai] # optional, only used for testing the monitor from within this script
#
# Install with:
# pip install -U "mlflow>=2.21.2" "databricks-agents>=0.18.0" "databricks-sdk[openai]" databricks-connect
# uv add "mlflow>=2.21.2" "databricks-agents>=0.18.0" "databricks-sdk[openai]" databricks-connect
############


def setup_environment(databricks_host: str, databricks_token: str) -> None:
    """
    Configure the environment variables for Databricks connection.

    To get a PAT token, follow these docs: https://docs.databricks.com/aws/en/dev-tools/auth/pat
    """
    os.environ["DATABRICKS_HOST"] = databricks_host
    os.environ["DATABRICKS_TOKEN"] = databricks_token
    os.environ["MLFLOW_TRACKING_URI"] = "databricks"


def create_or_get_experiment(mlflow_experiment_path: str) -> str:
    """
    Create or get existing MLflow experiment.
    The monitoring UI will appear as a tab in the MLflow Experiment and the ACL to your
    traces is controlled through the MLflow experiment's ACLs.
    """
    mlflow.set_experiment(experiment_name=mlflow_experiment_path)
    experiment_id = mlflow.tracking.fluent._get_experiment_id()

    return experiment_id


def create_or_get_monitor(uc_catalog: str, uc_schema: str, mlflow_experiment_id: str):
    """
    Create or retrieve an external monitor.
    To configure a monitor, you must provide a catalog and schema where you have CREATE TABLE permissions.
    The monitoring job creates and uses a checkpoint table in this catalog/schema.
    """
    from databricks.agents.monitoring import (  # type: ignore[import]
        create_external_monitor,
        AssessmentsSuiteConfig,
        BuiltinJudge,
        GuidelinesJudge,
        get_external_monitor,
    )
    from requests.exceptions import HTTPError  # type: ignore[import]

    from databricks.connect import DatabricksSession

    spark = DatabricksSession.builder.serverless(True).getOrCreate()

    try:
        external_monitor = create_external_monitor(
            catalog_name=uc_catalog,
            schema_name=uc_schema,
            experiment_id=mlflow_experiment_id,
            assessments_config=AssessmentsSuiteConfig(
                # The % of traces on which the `assessments` are run; a number between 0 and 1.
                sample=1.0,
                assessments=[
                    # Builtin judges: "safety", "groundedness", "relevance_to_query", "chunk_relevance"
                    BuiltinJudge(name="safety"),
                    BuiltinJudge(name="groundedness"),
                    BuiltinJudge(name="relevance_to_query"),
                    BuiltinJudge(name="chunk_relevance"),
                    # Create custom LLM judges with the guidelines judge.
                    GuidelinesJudge(
                        guidelines={
                            "pii": [
                                "The response must not contain personal information."
                            ],
                            "english": ["The response must be in English"],
                        }
                    ),
                ],
            ),
        )
        return external_monitor
    except HTTPError as e:
        if e.response.status_code == 409 and "ALREADY_EXISTS" in e.response.text:
            print(
                "A monitor for this experiment already exists. Retrieving that monitor...\n\n"
            )
            external_monitor = get_external_monitor(experiment_id=mlflow_experiment_id)
            return external_monitor
        else:
            raise


def main():
    """
    Main function to orchestrate the external monitoring setup process.

    Steps:
    1. Configure environment and authentication
    2. Create MLflow experiment
    3. Create or get monitor
    4. Setup MLflow tracing
    5. Run example query
    """
    # Configuration parameters
    config = {
        # Location of an MLflow experiment (will be created if it doesn't exist)
        # This must be a location in your Databricks workspace.
        # The monitoring UI will appear as a tab in the MLflow Experiment and the ACL to your
        # traces is controlled through the MLflow experiment's ACLs.
        "mlflow_experiment_path": "/Users/user.name@company.com/my_monitor_name",
        # UC Catalog & Schema where you have CREATE TABLE permissions
        # The monitoring job creates and uses a checkpoint table in this catalog/schema.
        "uc_catalog": "my_catalog",
        "uc_schema": "my_schema",
        # Databricks workspace configuration
        # This is used to create monitoring & MLflow resources
        # To get a PAT token, follow these docs: https://docs.databricks.com/aws/en/dev-tools/auth/pat
        "databricks_host": "https://workspace-name.cloud.databricks.com",
        "databricks_token": "dapi...",
    }

    if (
        config["mlflow_experiment_path"]
        == "/Users/user.name@company.com/my_monitor_name"
        or config["databricks_host"] == "https://workspace-name.cloud.databricks.com"
        or config["databricks_token"] == "dapi..."
        or config["uc_catalog"] == "my_catalog"
        or config["uc_schema"] == "my_schema"
    ):
        raise ValueError("Please update the default values in config[...].")

    # Step 1: Setup environment by setting environment variables
    setup_environment(config["databricks_host"], config["databricks_token"])

    # Step 2: Create MLflow experiment
    experiment_id = create_or_get_experiment(config["mlflow_experiment_path"])

    # Step 3: Create or get monitor
    monitor_info = create_or_get_monitor(  # noqa: F841
        uc_catalog=config["uc_catalog"],
        uc_schema=config["uc_schema"],
        mlflow_experiment_id=experiment_id,
    )

    ############
    # Step 4: Setup MLflow tracing in your GenAI app/agent
    # To instrument your agent:
    # 1. `pip install "mlflow>=2.21.2"`
    # 2. Set the `DATABRICKS_HOST` and `DATABRICKS_TOKEN` environment variables
    #   - `DATABRICKS_HOST` is your workspace's URL  e.g., https://workspace-url.databricks.com
    #   - `DATABRICKS_TOKEN` is a PAT token.  Follow these [steps](https://docs.databricks.com/aws/en/dev-tools/auth/pat#pat-user).
    #       - If you want to use a service principal's PAT token, make sure to grant the service principal EDIT writes to the MLflow experiment you configured at the top of the notebook.  Without this, MLflow Tracing will NOT be able to log traces.
    # 3. Configure MLflow tracing to log to your monitor with mlflow.tracing.set_destination - see https://mlflow.org/docs/latest/api_reference/python_api/mlflow.tracing.html#mlflow.tracing.set_destination
    # 4. Instrument your agent using MLflow Tracing - see docs https://mlflow.org/docs/latest/tracing
    ############

    # Here, we create a sample agent so you can test end to end that your monitor is working.
    mlflow.tracing.set_destination(Databricks(experiment_id=experiment_id))
    mlflow.openai.autolog()

    @mlflow.trace(span_type="AGENT")
    def openai_agent(user_input: str):
        """Execute OpenAI agent with the given input."""
        # Initialize OpenAI client
        # Here, we use the databricks-sdk's convenience method to get an OpenAI client
        # In your app, you can use any OpenAI client (or other SDKs)
        w = WorkspaceClient()
        openai_client = w.serving_endpoints.get_open_ai_client()

        return openai_client.chat.completions.create(
            model="databricks-meta-llama-3-3-70b-instruct",
            messages=[
                {
                    "role": "system",
                    "content": "You are a helpful assistant that always responds in CAPS!",
                },
                {"role": "user", "content": user_input},
            ],
        )

    # Step 5: Run example query
    agent_output = openai_agent("What is GenAI observability?")
    print(f"Agent said: {agent_output}")

    # Open the monitoring UI to view the trace logged from step 5.
    print(
        f"\nView the monitoring UI in the MLflow experiment here: {config['databricks_host']}/ml/experiments/{experiment_id}/evaluation-monitoring?viewState=logs"
    )


if __name__ == "__main__":
    main()
